"""
@file collect_data.py

@brief this file is responsible for collecting RF background levels

At a high-level:

With some frequency defined in a config file:
- Sample the current signal coming in from the line-in or an RTLSDR
 (we assume this is a radio)
- Translate to S-units (extended s-unit scale)
- Collect lat, lng, elevation
- Create a row in an output CSV
- Announce out loud the S-level so the user can slow down

This assumes you have set up an actual radio as an input to the computer, and it works correctly. In future, an example setup will be added to the README
"""

# https://stackoverflow.com/questions/40138031/how-to-read-realtime-microphone-audio-volume-in-python-and-ffmpeg-or-similar (mic audio level)
# https://python.plainenglish.io/receiving-and-processing-gps-data-using-external-receiver-with-python-24d3592ad2e0 (gps recv)

import platform
import time
import atexit
import numpy as np
import subprocess
import logging
import csv
import os
import datetime
import sys

from rtlsdr import RtlSdr
import serial
from serial.tools import list_ports

from kermit_sdr.config import SAMPLE_INTERVAL, SignalSource, SIGNAL_SOURCE, RtlSdrSettings, ANTENNA_FUDGE_FACTOR,\
                   LISTENING_FREQUENCY, ANNOUNCE_SIGNAL, ANNOUNCE_SIGNAL_EVERY, S_UNIT_SCALE, GPS_POLL_SEC,\
                   OUTPUT_FILE_CSV, DEDUPLICATE
from kermit_sdr.utils import get_platform_speak_func, GpsReadQuality, GpsResponse
"""
If we don't find a GPS, still announce signal but don't store location data
"""
store_location = True
"""
The shape of the output CSV file
"""
CSV_COLUMNS = [
    "timestamp", "signal_strength_db", "signal_strength_s_unit", "latitude",
    "longitude", "elevation"
]


def get_s_unit_from_db(signal_strength_db):
    """
    Get the S-unit of a signal given it's strength in dB

    @param signal_strength_db: The signal strength in dB
    """
    for key, interval in S_UNIT_SCALE.items():
        if interval.in_interval(signal_strength_db):
            logging.debug(
                f"Signal strength {signal_strength_db} corresponds to {key}")
            return key
    return "Unknown S-unit"


def dbfft(x, fs, win=None, ref=1):
    """
    Calculate spectrum in dB scale
    Args:
        x: input signal
        fs: sampling frequency
        win: vector containing window samples (same length as x).
             If not provided, then rectangular window is used by default.
        ref: reference value used for dBFS scale. 32768 for int16 and 1 for float

    Returns:
        freq: frequency vector
        s_db: spectrum in dB scale
    """

    N = len(x)  # Length of input sequence

    if win is None:
        win = np.ones([N, 1])
    if len(x) != len(win):
        raise ValueError('Signal and window must be of the same length')
    x = x * win

    # Calculate real FFT and frequency vector
    sp = np.fft.rfft(x)
    freq = np.arange((N / 2) + 1) / (float(N) / fs)

    # Scale the magnitude of FFT by window and factor of 2,
    # because we are using half of FFT spectrum.
    s_mag = np.abs(sp) * 2 / np.sum(win)

    # Convert to dBFS
    s_dbfs = 20 * np.log10(s_mag / ref)

    return freq, s_dbfs


def read_signal_str(sdr):
    """
    Read the signal strength out from our signal source

    @param sdr - the RTLSDR object, if we're using the RTLSDR
                 None if we're not
    """
    if sdr:
        # this is TOO EASY
        logging.debug(
            f"Reading signal samples from RTLSDR at sample rate {RtlSdrSettings.sample_rate}"
        )
        samples = sdr.read_samples(1024)
    else:
        #collect_samples_from_line_in()
        pass

    # Once we have some samples of the signal, we can extract the power
    freq, dbfs = dbfft(samples, RtlSdrSettings.sample_rate)
    logging.debug(
        f"Signal sample summary: min - {np.min(dbfs)}, max: - {np.max(dbfs)}, average - {np.average(dbfs)}"
    )
    # TODO - instead of doing the max, we really should specifically take the listening frequency
    # Or something like the range of the antenna's useful frequencies (requires user config input)
    # But for almost all practical intents and purposes, the max _is_ the listening frequency
    max_db = np.max(dbfs)
    logging.debug(
        f"Shifting measured signal {max_db} by antenna gain {ANTENNA_FUDGE_FACTOR}"
    )
    db_result = max_db - ANTENNA_FUDGE_FACTOR
    return db_result


def announce_signal(signal_strength_s_unit, speak_func):
    """
    Announce the signal strength using system text to speech
    
    @param signal_strength_db: Signal strength in dB
    @param speak_func: speak function to use for announcin strength
    """
    speak_func(signal_strength_s_unit)


def get_gga_signal_from_serial(serial_obj):
    """
    Given a GPS serial object, decode the GGA signal, and return relevant information

    @param serial_obj - a serial object that is a GPS unit
    @return - timestamp, lat, lng, quality, number of satellites, altitude
    """
    line = serial_obj.readline().decode('utf-8')
    # Only use GPGGA signal type
    line_found = False
    while not line_found:
        if "GPGGA" in line:
            logging.debug("Found GPGGA signal from GPS unit")
            line_found = True
            line_split = line.split(",")
            if int(line_split[6]) == 0:
                return None
            break
        line = serial_obj.readline().decode('utf-8')

    # Process the GPGGA signal, return the serial object
    # The GPGGA signal is well-defined (for decades),
    # so we can be a bit hacky here and get away with it.
    line_split = line.split(",")
    res = GpsResponse(timestamp=line_split[1],
                      quality=GpsReadQuality(int(line_split[6])),
                      sat_count=int(line_split[7]),
                      altitude=float(line_split[9]),
                      error=float(line_split[8]))

    # Converting the degrees from NMEA to decimal is kind of annoying
    # The format is "ddmm.mmm" for latitude, and
    # "dddmm.mmmmm" for longitude
    # Direction is encoded in the next entry
    lat = line_split[2]
    lat_dir = line_split[3]
    lng = line_split[4]
    lng_dir = line_split[5]
    lat_deg = lat[0:2]
    lat_min = lat[2:]
    lng_deg = lng[0:3]
    lng_min = lng[3:]

    latitude = float(lat_deg) + (float(lat_min) / 60)
    longitude = float(lng_deg) + (float(lng_min) / 60)
    if lat_dir == "S":
        latitude *= -1
    if lng_dir == "W":
        longitude *= -1

    # Relevant XKCD - https://xkcd.com/2170/
    res.lat = format(latitude, ".5f")
    res.lng = format(longitude, ".5f")
    logging.info(f"Read GGA signal {res} successfully")

    return res


def setup_gps_source():
    """
    Find a GPS source over serial and return something that we can read from

    @return a Serial object that can be read for GPS data
    """
    port_list = list(list_ports.comports())
    logging.debug(
        f"Candidate GPS devices: [{', '.join([p.device for p in port_list])}]")

    # If it tells us it is a GPS, we're pretty sure it is a GPS
    found_gps_for_sure = False
    for p in port_list:
        if "gps" in str(p.product) or "GPS" in str(p.product):
            port = p.device
            found_gps_for_sure = True
            break

    # Otherwise, if there's more than one option, ask the user
    if len(port_list) > 1 and not found_gps_for_sure:
        port_str = "\n\t"
        port_str += "\n\t".join([p.device for p in port_list])
        inp = input(
            f"Which port corresponds to your GPS? Options: {port_str}\n $")
        for port_obj in port_list:
            if port_obj.device == inp:
                port = port_obj.device
                break

    elif len(port_list) == 1:
        # If there's only one option, we will try it because it's all there is.
        port = port_list[0].device
    elif not found_gps_for_sure:
        # Otherwise, we fail
        logging.warning("Could not find GPS. Not logging location source")
        return None
    logging.debug(f"Found GPS source device {port}")
    serial_obj = serial.Serial(port)
    # Wait for it to start returning location data
    for _ in range(GPS_POLL_SEC):
        gps_response = get_gga_signal_from_serial(serial_obj)
        if gps_response is None:
            continue
        if gps_response.is_valid_read():
            logging.info(f"Found properly functioning GPS device at {port}")
            return serial_obj
        time.sleep(1)

    # If we get here, we've tried for the max timeout. Give up
    logging.error(
        f"Polled for {GPS_POLL_SEC} seconds but could not get valid signal from GPS. No location services available"
    )
    return None


def save_read_to_file(expanded_out_file, signal_strength_db,
                      signal_strength_s_unit, gps_read):
    """
    Save a single sample to the CSV file. Assumes we've set it up already
    and there is a correctly populated header.
    """
    row_dict = {}
    # set up dictionary for dictwriter
    for col in CSV_COLUMNS:
        row_dict[col] = ""

    row_dict["timestamp"] = str(datetime.datetime.now())

    [
        "timestamp", "signal_strength_db", "signal_strength_s_unit",
        "latitude", "longitude", "elevation"
    ]
    # Populate GPS reads if the GPS read is a valid object
    if gps_read is not None:
        row_dict["latitude"] = gps_read.lat
        row_dict["longitude"] = gps_read.lng
        row_dict["elevation"] = gps_read.altitude

    row_dict["signal_strength_db"] = signal_strength_db
    row_dict["signal_strength_s_unit"] = signal_strength_s_unit

    # Open in append mode
    with open(expanded_out_file, "a") as fh:
        writer = csv.DictWriter(fh, fieldnames=CSV_COLUMNS)
        writer.writerow(row_dict)


def init_output_file(expanded_out_file):
    """
    Write the CSV header to the output file
    """
    new_file = True

    if os.path.exists(expanded_out_file):
        new_file = False
        logging.warning(f"Output file {expanded_out_file} exists already")
        response = input(
            "Do you want to overwrite the existing output file? y/N\n")
        if response == "y" or response == "Y":
            new_file = True

    # We overwrite the file if it doesn't exist, or if the user told us to.
    if not new_file:
        return False

    # open in "w" mode to write a new file
    with open(expanded_out_file, "w") as fh:
        writer = csv.writer(fh)
        writer.writerow(CSV_COLUMNS)


def main(args=None):
    """
	Main program loop.

    @param args: arguments from CLI entry point.
        Only supported argument is the output file
	"""

    # Decide if the user has supplied their own output file over CLI
    output_file_base = OUTPUT_FILE_CSV
    if args is not None:
        if args.output_file is not None and args.output_file != "":
            output_file_base = f"{args.output_file}.csv"

    # Detect the platform
    system = platform.system()
    logging.debug(f"Found platform {system}")
    speak_func = get_platform_speak_func(system)

    expanded_out_file = os.path.expandvars(
        os.path.expanduser(output_file_base))

    gps_stream = setup_gps_source()

    if gps_stream == None:
        logging.warning("GPS source not found. Not storing location data")
        store_location = False

    # Leave this obj as None if we are not using the RTLSDR
    sdr = None

    response = init_output_file(expanded_out_file)
    if response == False:
        logging.error(
            "Failed to create output file. Please verify your supplied output file path is free, or you are okay overwriting it"
        )
        sys.exit(1)

    if SIGNAL_SOURCE == SignalSource.RTLSDR:
        # Configure the RTLSDR if we want it
        logging.debug("Using signal source RTLSDR")
        sdr = RtlSdr()
        if LISTENING_FREQUENCY <= 3e7:
            # For HF frequencies, we will get better results out of direct sampling
            sdr.set_direct_sampling(True)
        sdr.sample_rate = RtlSdrSettings.sample_rate
        sdr.center_freq = RtlSdrSettings.center_freq
        sdr.freq_correction = RtlSdrSettings.freq_correction
        sdr.gain = RtlSdrSettings.gain
        atexit.register(sdr.close)
    else:
        logging.debug("Using signal source LINE IN")
        logging.warning("LINE IN signal source is not yet supported")
        sys.exit(0)

    # Iteration counter. Used for announcing the signal strength every however often
    i = 0
    while True:

        # Get and process the signal strength
        signal_strength_db = read_signal_str(sdr)
        signal_strength_s_unit = get_s_unit_from_db(signal_strength_db)
        logging.info(
            f"Signal strength {format(signal_strength_db, '.3f')} dB ({signal_strength_s_unit})."
        )

        gps_read = None

        # Get and process the location
        if gps_stream is not None:
            gps_read = get_gga_signal_from_serial(gps_stream)

        save_read_to_file(expanded_out_file, signal_strength_db,
                          signal_strength_s_unit, gps_read)

        time.sleep(SAMPLE_INTERVAL)
        if ANNOUNCE_SIGNAL and (i % ANNOUNCE_SIGNAL_EVERY == 0):
            logging.debug(
                f"Announcing signal strength {signal_strength_db} dB ({signal_strength_db})"
            )
            announce_signal(signal_strength_s_unit, speak_func)

        i += 1


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    main()



"""
@file config.py

@brief This file maintains configuration details for the program

"""

from enum import Enum, auto
from kermit_sdr.utils import S_UNIT_SCALE_HF, S_UNIT_SCALE_VHF


class SignalSource(Enum):
    """
	This class represents a signal source that the program should read the signal level from
	"""
    # A USB RTL-SDR
    # https://www.rtl-sdr.com
    RTLSDR = auto()
    # A microphone plugged into line-in
    # For mac, we assume the headphone port is being used for line-in
    # For others, line in is assumed
    # Note - line-in will only ever work for AM signals. Make sure your radio is set to AM mode.
    LINE_IN = auto()
    # TODO more like USB line in etc


"""
The frequency to listen on, in Hz

A sensible default is 146.52 (MHz), as that is an often-empty conventional simplex frequency.

It's usually pretty easy to make test transmissions on that frequency

Set it to something that plays nice with your antenna and radio.
"""
LISTENING_FREQUENCY = 145230000
"""
Because s-unit scales are different for VHF vs HF, we need a config parameter to specify to the program whether the user is listening in VHF or HF

Default is set based on listening frequency but can be overridden if necessary
"""
S_UNIT_SCALE = S_UNIT_SCALE_VHF if LISTENING_FREQUENCY >= 3e7 else S_UNIT_SCALE_HF


class RtlSdrSettings:
    """
    Class that stores RTL SDR settings for initialization
    """
    sample_rate = 2.048e6  # Sample rate in Hz
    center_freq = LISTENING_FREQUENCY  # Center frequency, in Hz
    freq_correction = 1  # Parts per million
    gain = 0  # Be careful when overriding this.


"""
Because the antenna is not an isotropic radiator, and the input is all in relative db, not absolute db, we need to be able to convert to actual dbi.

We assume you are using a ham-stick or similar simple antenna, by default, which basically behaves like a 0 dBi antenna.
"""
ANTENNA_FUDGE_FACTOR = 0
"""
Set the program's signal source.
Today, RTLSDR and LINE IN are supported
Defaults to RTLSDR

Note that LINE IN only will work if set to recieve in AM mode.
FM simply will not work as FM signal strength is not correlated to audio volume.
"""
SIGNAL_SOURCE = SignalSource.RTLSDR
"""
The /dev path that the USB serial device shows up at

Only used if we don't find it ourselves.
"""
GPS_DEV_PATH = ""
"""
Set the sampling interval, in seconds
"""
SAMPLE_INTERVAL = 1.0
"""
A flag used to enable/disable the annunciation of signal strength
"""
ANNOUNCE_SIGNAL = True
"""
Every _n_ samples, announce the signal
"""
ANNOUNCE_SIGNAL_EVERY = 5
"""
The output file path.

Only set OUTPUT_FILE_BASE. The OUTPUT_FILE_CSV and OUTPUT_FILE_MAP will be computed.

The program will ask you before overwriting, but won't automatically create new filenames.

If you provide an output file via the `--output-file` option, it will override this
"""
OUTPUT_FILE_BASE = "~/Desktop/kermit_output/test"
OUTPUT_FILE_CSV = f"{OUTPUT_FILE_BASE}.csv"
OUTPUT_FILE_MAP = f"{OUTPUT_FILE_BASE}.html"
"""
Should we auto-open the map in your browser?
"""
AUTO_OPEN_MAP = True
"""
How long in seconds should we wait before giving up on using GPS?
"""
GPS_POLL_SEC = 30
"""
De-duplicate consecutive entries from the same location
This is useful for things like getting stuck at a red light, where you end up lighting up the heat maps,
because there are too many samples in one place

Note that this process happens at map generation time and we will not remove data points from the CSV files.
"""
DEDUPLICATE = True
"""
The step size, in decimal degrees, to use for deduplication binning.
.001 is about 100 meters
.0001 is about 11 meters
.00001 is about 1.1 meters
"""
DEDUPLICATION_STEP = 0.0001



"""
@file sample_signal_distribution.py

@brief collect data from RTLSDR source for mapping dB (RF) to dB (audio) 


The relationship between dB (of RF signal) and dB (of audio out of the line-in) may not be linear (in fact, it probably isn't linear)

There might be a better theorecitcal way to handle that mapping, and I promise to explore that in the future, but for now, what this file helps with is using the RTLSDR's ability to sample both, and produce a sample of the __shape__ of the function across frequencies and volume levels. We will then shift it by the signal strength added by the radio and microphone's relative volume/gain, and get a close-ish representation of the dB of the signal given the line-in.

NOTE: this only has a chance in hell of working on AM signals. Make sure your radio is in AM mode.
"""



#!/usr/bin/env python3
"""
This is the entry point to KERMIT. It can be used to collect data and generate maps.

For more details about KERMIT, please see https://github.com/benhg/kermit#readme
"""

import argparse
import logging
import sys

from kermit_sdr.generate_map import main as map_main
from kermit_sdr.collect_data import main as collect_main
from kermit_sdr.version import VERSION


def main():
    """
    The main entry point

    @param args: The arguments namespace from the parser
    """
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "action",
        metavar='action',
        type=str,
        help=
        "The action for KERMIT to take. Current options: ['generate-map', 'collect-data', 'version'']",
        choices=["generate-map", "collect-data", "version"])
    parser.add_argument(
        "-o",
        "--output-file",
        type=str,
        required=False,
        help=
        "Provide an output file (for data collection) or both an input and an output file (for map generation). Provide a filename with no extension as KERMIT will add appropriate extensions. This output file overrides the one in the config file."
    )
    args = parser.parse_args()

    if args.action == 'version':
        print(f"KERMIT version {VERSION}")
    elif args.action == "generate-map":
        map_main(args=args)
    elif args.action == "collect-data":
        collect_main(args=args)


if __name__ == '__main__':
    main()



"""
@file generate_map.py

@brief This file is responsible for generating maps based on CSVs created by the `collect_data.py` file.

Taking a CSV of locations and signal strengths, overlay that onto a map using OpenStreetMap

Heavy inspiration taken from https://python-charts.com/spatial/spatial-heatmap-plotly/
"""

import os
import logging

import pandas as pd
import numpy as np
import plotly.express as px
import plotly

from kermit_sdr.config import OUTPUT_FILE_CSV, OUTPUT_FILE_MAP, AUTO_OPEN_MAP, DEDUPLICATE, DEDUPLICATION_STEP


def main(args=None):
    """
    Main program entry point

    @param args: arguments from CLI entry point.
        Only supported argument is the output file
    """

    # Decide if the user has supplied their own output file over CLI
    output_file_csv = OUTPUT_FILE_CSV
    output_file_map = OUTPUT_FILE_MAP
    if args is not None:
        if args.output_file is not None and args.output_file != "":
            output_file_map = f"{args.output_file}.html"
            output_file_csv = f"{args.output_file}.csv"

    # Data with latitude/longitude and values
    csv_file = os.path.expandvars(os.path.expanduser(output_file_csv))
    map_file = os.path.expandvars(os.path.expanduser(output_file_map))
    df = pd.read_csv(csv_file)
    logging.info(f"Found {len(df['latitude'])} data points")
    if DEDUPLICATE:
        # Bin all the points into groups of 11-meter segments (4 decimal digits)
        # Take the max signal strength of each bin
        # Those are our deduplicated data points
        summary_points = []
        step = DEDUPLICATION_STEP
        to_bin = lambda x: np.floor(x / DEDUPLICATION_STEP
                                    ) * DEDUPLICATION_STEP
        df["lat_bin"] = to_bin(df.latitude)
        df["lon_bin"] = to_bin(df.longitude)
        groups = df.groupby(["lat_bin", "lon_bin"])
        for group in groups:
            lat = group[0][0]
            lon = group[0][1]
            signal_str_db = np.max(group[1].signal_strength_db)
            summary_points.append({
                "latitude": lat,
                "longitude": lon,
                "signal_strength_db": signal_str_db
            })

        df = pd.DataFrame(summary_points)
        logging.info(
            f"Found {len(df['latitude'])} data points after deduplication")

    fig = px.density_mapbox(
        df,
        lat='latitude',
        lon='longitude',
        z='signal_strength_db',
        radius=8,
        # Center at the first entry
        center=dict(lat=df["latitude"][1], lon=df["longitude"][1]),
        zoom=10,
        mapbox_style='open-street-map')

    logging.info(f"Saving map to {map_file}")
    plotly.offline.plot(fig, filename=map_file, auto_open=AUTO_OPEN_MAP)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    main()



"""
@file setup.py
@brief the PyPI required setup.py
"""

import os
from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

version_ns = {}
with open(os.path.join("kermit_sdr", "version.py")) as f:
    exec(f.read(), version_ns)
version = version_ns['VERSION']

with open('requirements.txt') as f:
    install_requires = f.readlines()

setup(name='kermit-sdr',
      version=version,
      license='MIT',
      author="Benjamin Glick",
      author_email='glick@glick.cloud',
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: Apache Software License",
          "Natural Language :: English", "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Topic :: Scientific/Engineering"
      ],
      python_requires=">=3.6.0",
      packages=find_packages(),
      url='https://github.com/benhg/kermit',
      keywords=["kermit", "radio", "SDR", "map", "geospatial"],
      entry_points={'console_scripts': ['kermit=kermit_sdr.kermit:main']},
      install_requires=install_requires,
      long_description=long_description,
      long_description_content_type='text/markdown'
)



"""
@file utils.py

@brief this file is responsible for maintaining functions and data structures that are used across multiple files
"""
"""
Map dB ranges to S-units, for clear annunciation
"""
import subprocess
from enum import IntEnum


class Interval():
    """
    This class represents a range between two numbers
    We use it for our S-unit scales
    """
    def __init__(self, start=0, end=0):
        """
        Constructor for interval class
        """
        self.start = start
        self.end = end

    def in_interval(self, val):
        """
        Check if a number is in this interval
        """
        return (val > self.start and val <= self.end)


"""
The S-unit scales

To make the announcments shorter, we use an extended scale, so above S-9,
    instead of saying S-9 plus x, we instead define S_x as S_(x-1) + 6 dB

We also add S-0 for below -48
"""

S_UNIT_SCALE_HF = {
    "S0": Interval(start=-999, end=-48),
    "S1": Interval(start=-48, end=-42),
    "S2": Interval(start=-42, end=-36),
    "S3": Interval(start=-36, end=-30),
    "S4": Interval(start=-30, end=-24),
    "S5": Interval(start=-24, end=-18),
    "S6": Interval(start=-18, end=-12),
    "S7": Interval(start=-12, end=-6),
    "S8": Interval(start=-6, end=0),
    "S9": Interval(start=0, end=6),
    "S10": Interval(start=6, end=12),
    "S11": Interval(start=12, end=18),
    "S12": Interval(start=18, end=24),
    "S too much": Interval(start=24, end=999)
}

S_UNIT_SCALE_VHF = {
    "S0": Interval(start=-999, end=-48),
    "S1": Interval(start=-48, end=-42),
    "S2": Interval(start=-42, end=-36),
    "S3": Interval(start=-36, end=-30),
    "S4": Interval(start=-30, end=-24),
    "S5": Interval(start=-24, end=-18),
    "S6": Interval(start=-18, end=-12),
    "S7": Interval(start=-12, end=-6),
    "S8": Interval(start=-6, end=0),
    "S9": Interval(start=0, end=6),
    "S10": Interval(start=6, end=12),
    "S11": Interval(start=12, end=18),
    "S12": Interval(start=18, end=24),
    "S too much": Interval(start=24, end=999)
}


def speak_darwin(text):
    """
    The function for using the `say` command on MacOS
    """
    subprocess.Popen(["say", f"'{text}'"])


def speak_linux(text):
    """
    The function for using the `espeak` command on Linux
    """
    subprocess.Popen(["echo", f"'{text}'", "|", "espeak"])


def get_platform_speak_func(platform):
    """
    Get a callable that uses the system's text-to-speech program to say a word.

    Returns a function that has one parameter, the text to say
    """
    if platform == "Darwin":
        return speak_darwin
    elif platform == "Linux":
        return speak_linux


class GpsReadQuality(IntEnum):
    """
    GPS read quality (source)

    IntEnum because these have well-defined values
    So we can cast from an integer freely

    See https://www.oc.nps.edu/oc2902w/gps/gpsacc.html for some details
    """
    INVALID = 0
    GPS_SPS = 1
    DGPS = 2
    PPS = 3
    REAL_TIME_KINEMATIC = 4
    FLOAT_RTK = 5
    ESTIMATED = 6
    MANUAL_INPUT = 7
    SIMULATED = 8


class GpsResponse:
    """
    A read from a GPS GGA signal. Contains the following attributes:

    Details from here: http://www.gpsinformation.org/dale/nmea.htm#GGA

    @var timestamp - the timestamp associated with the read
    @var lat - latitude
    @var lng - longitude
    @var quality - enum representing read quality
    @var sat_count - number of satellites currently being read.
    @var altitude - the GPS altitude
    @var error - the dilution of precision - above 20 == bad.
    @var timestamp - timestamp of read
    """
    def __init__(self,
                 lat=0,
                 lng=0,
                 quality=GpsReadQuality.INVALID,
                 sat_count=0,
                 altitude=0,
                 error=0,
                 timestamp=0):
        """
        Constructor for GPS response class
        """
        self.lat = lat
        self.lng = lng
        self.quality = quality
        self.sat_count = sat_count
        self.altitude = altitude
        self.error = error
        self.timestamp = timestamp

    def is_valid_read(self):
        """
        Return whether this read is trustworthy
        """
        return self.quality != GpsReadQuality.INVALID and self.error < 20 and self.sat_count >= 3

    def __repr__(self):
        return f"GpsResponse<lat={self.lat}, lng={self.lng}, quality={GpsReadQuality(self.quality).name}, sat_count={self.sat_count}, altitude={self.altitude}, error={self.error}, timestamp={self.timestamp}>"
